import numpy as np
import ctypes
from scipy.sparse import issparse, isspmatrix_csr
import warnings
from sklearn.base import BaseEstimator
from . import _wrapper_double, _wrapper_float

__all__ = ["LinearRegression", "ElasticNet"]

def _is_row_major(X):
    if (X is None) or (issparse(X)):
        return True
    else:
        return X.strides[1] == X.dtype.itemsize

def _copy_if_subview(X):
    ### TODO: the Cython functions should accept a 'leading dimension'
    ### parameter so as to avoid copying the data here
    if (X is not None) and (not issparse(X)):
        row_major = _is_row_major(X)
        leading_dimension = int(X.strides[0 if row_major else 1] / X.dtype.itemsize)
        if leading_dimension != X.shape[1 if row_major else 0]:
            X = X.copy()
        elif (len(X.strides) != 2) or (not X.flags.aligned):
            X = X.copy()
        if _is_row_major(X) != row_major:
            X = np.ascontiguousarray(X)
    return X

def _copy_if_subview_1d(arr):
    if arr is not None:
        if not isinstance(arr, np.ndarray):
            arr = np.ascontiguousarray(arr)
        arr = arr.reshape(-1)
        if (arr.strides[0] != arr.dtype.itemsize) or (not arr.flags.aligned):
            arr = np.ascontiguousarray(arr)
    return arr

class LinearRegression(BaseEstimator):
    """
    Linear Regression
    
    Typical Linear Regression model, which keeps track of the aggregated data
    needed to obtain the closed-form solution in a way that calling 'partial_fit'
    multiple times would be equivalent to a single call to 'fit' with all the data.
    This is an exact method rather than a stochastic optimization procedure.

    Also provides functionality for making predictions according to upper confidence
    bound (UCB) and to Thompson sampling criteria.

    Note
    ----
    Doing linear regression this way requires both memory and computation time
    which scale quadratically with the number of columns/features/variables. As
    such, the class will by default use C 'float' types (typically ``np.float32``)
    instead of C 'double' (``np.float64``), in order to save memory.

    Parameters
    ----------
    lambda_ : float
        Strenght of the L2 regularization.
    fit_intercept : bool
        Whether to add an intercept term to the formula. If passing 'True', it will
        be the last entry in the coefficients.
    method : str, one of 'chol' or 'sm'
        Method used to fit the model. Options are:

        ``'chol'``:
            Uses the Cholesky decomposition to solve the linear system from the
            least-squares closed-form each time 'fit' or 'partial_fit' is called.
            This is likely to be faster when fitting the model to a large number
            of observations at once, and is able to better exploit multi-threading.
        ``'sm'``:
            Starts with an inverse diagonal matrix and updates it as each
            new observation comes using the Sherman-Morrison formula, thus
            never explicitly solving the linear system, nor needing to calculate
            a matrix inverse. This is likely to be faster when fitting the model
            to small batches of observations. Be aware that with this method, it
            will add regularization to the intercept if passing 'fit_intercept=True'.

        Note that it is possible to change the method after the object has
        already been fit (e.g. if you want non-regularization intercept
        with fast online updates, you might use Cholesky first and then switch
        to Sherman-Morrison).
    calc_inv : bool
        When using ``method='chol'``, whether to also produce a matrix inverse, which
        is required for using the LinUCB prediction mode. Ignored when
        passing ``method='sm'`` (the default). Note that is is possible to change
        the method after the object has already been fit.
    precompute_ts : bool
        Whether to pre-compute the necessary matrices to accelerate the Thompson
        sampling prediction mode (method ``predict_thompson``). If you plan to use
        ``predict_thompson``, it's recommended to pass "True".
        Note that this will make the Sherman-Morrison updates (``method="sm"``)
        much slower as it will calculate eigenvalues after every update.
        Can be changed after the object is already initialized or fitted.
    precompute_ts_multiplier : float
        Multiplier for the covariance matrix to use when using ``precompute_ts``.
        Calling ``predict_thompson`` with this same multiplier will be faster than
        with a different one. Calling it with a different multiplier with
        ``precompute_ts`` will still be faster than without it, unless using
        also ``n_presampled``.
        Ignored when passing ``precompute_ts=False``.
    n_presampled : None or int
        When passing ``precompute_ts``, this denotes a number of coefficients to pre-sample
        after calling 'fit' and/or 'partial_fit', which will be used later
        when calling ``predict_thompson`` with the same multiplier as in ``precompute_ts_multiplier``.
        Pre-sampling a large number of coefficients can help to speed up Thompson-sampled predictions
        at the expense of longer fitting times, and is recommended if there is a large number of
        predictions between calls to 'fit' or 'partial_fit'.
        If passing 'None' (the default), will not pre-sample a finite number of the coefficients
        at fitting time, but will rather sample (different) coefficients in calls to
        ``predict_thompson``.
        The pre-sampled coefficients will not be used if calling ``predict_thompson`` with
        a different multiplier than what was passed to ``precompute_ts_multiplier``.
    rng_presample : None, int, RandomState, or Generator
        Random number generator to use for pre-sampling coefficients.
        If passing an integer, will use it as a random seed for initialization. If passing
        a RandomState, will use it to draw an integer to use as seed. If passing a
        Generator, will use it directly. If passing 'None', will initialize a Generator
        without random seed.
        Ignored if passing ``precompute_ts=False`` or ``n_presampled=None`` (the defaults).
    use_float : bool
        Whether to use C 'float' type for the required matrices. If passing 'False',
        will use C 'double'. Be aware that memory usage for this model can grow
        very large. Can be changed after initialization.

    Attributes
    ----------
    coef_ : array(n) or array(n+1)
        The obtained coefficients. If passing 'fit_intercept=True', the intercept
        will be at the last entry.
    """
    def __init__(self, lambda_=1., fit_intercept=True, method="sm",
                 calc_inv=True, precompute_ts=False,
                 precompute_ts_multiplier=1.,
                 n_presampled=None, rng_presample=None,
                 use_float=True):
        self.lambda_ = lambda_
        self.fit_intercept = fit_intercept
        self._method = method
        self._calc_inv = bool(calc_inv)
        self._precompute_ts = bool(precompute_ts)
        self._precompute_ts_multiplier = precompute_ts_multiplier
        self._n_presampled = n_presampled
        self.rng_presample = rng_presample
        self._use_float = bool(use_float)

        self._set_dtype(force_cast=False)
        self.coef_ = np.empty(0, dtype=self._dtype)
        self._XtX = np.empty((0,0), dtype=self._dtype)
        self._invXtX = np.empty((0,0), dtype=self._dtype)
        self._XtY = np.empty(0, dtype=self._dtype)
        self._bufferX = np.empty(0, dtype=self._dtype)
        self._n = 0

        ### These are only used alongside precomputed TS
        self._EigMultiplier = np.empty((0,0), dtype=self._dtype)
        self._EigValsUsed = np.empty(0, dtype=self._dtype)
        self._EigValsOrig = np.empty(0, dtype=self._dtype)
        self._coef_precomputed = np.empty((0,0), dtype=self._dtype)

        self.is_fitted_ = False

    def _set_dtype(self, force_cast=False):
        self._dtype = ctypes.c_float if self._use_float else ctypes.c_double

        if force_cast:
            if self.coef_.dtype != self._dtype:
                self.coef_ = self.coef_.astype(self._dtype)
            if self._XtX.dtype != self._dtype:
                self._XtX = self._XtX.astype(self._dtype)
            if self._invXtX.dtype != self._dtype:
                self._invXtX = self._invXtX.astype(self._dtype)
            if self._XtY.dtype != self._dtype:
                self._XtY = self._XtY.astype(self._dtype)
            if self._bufferX.dtype != self._dtype:
                self._bufferX = self._bufferX.astype(self._dtype)
            if self._EigMultiplier.dtype != self._dtype:
                self._EigMultiplier = self._EigMultiplier.astype(self._dtype)
            if self._EigValsUsed.dtype != self._dtype:
                self._EigValsUsed = self._EigValsUsed.astype(self._dtype)
            if self._EigValsOrig.dtype != self._dtype:
                self._EigValsOrig = self._EigValsOrig.astype(self._dtype)
            if self._coef_precomputed.dtype != self._dtype:
                self._coef_precomputed = self._coef_precomputed.astype(self._dtype)

    @property
    def use_float(self):
        return self._use_float
    @use_float.setter
    def use_float(self, value):
        self._use_float = use_float
        self._set_dtype(force_cast=True)

    @property
    def method(self):
        return self._method
    
    @method.setter
    def method(self, value):
        assert value in ["chol", "sm"]
        if (self._method != value) and (self.is_fitted_):
            cy_funs = _wrapper_float if self._use_float else _wrapper_double
            if self._method == "sm":
                self._XtX = np.empty(self._invXtX.shape, dtype=self._dtype)
                cy_funs.get_matrix_inv(self._invXtX, self._XtX)
            elif (not self._calc_inv):
                self._invXtX = np.empty(self._XtX.shape, dtype=self._dtype)
                cy_funs.get_matrix_inv(self._XtX, self._invXtX)
                self._calc_inv = True
        self._method = value

    @property
    def calc_inv(self):
        return self._calc_inv
    
    @calc_inv.setter
    def calc_inv(self, value):
        if bool(value) != bool(self._calc_inv):
            if self._method == "chol":
                self._invXtX = np.empty(self._XtX.shape, dtype=self._dtype)
                if self.is_fitted_:
                    cy_funs = _wrapper_float if self._use_float else _wrapper_double
                    cy_funs.get_matrix_inv(self._XtX, self._invXtX)
            self._calc_inv = bool(value)

    @property
    def precompute_ts(self):
        return self._precompute_ts
    
    @precompute_ts.setter
    def precompute_ts(self, value):
        if (bool(value) != bool(self._precompute_ts)) and (self.is_fitted_):
            ### Do not set the value right away as the procedure can
            ### be interrupted in the middle and would leave the object
            ### in an unusable state
            if not value:
                self._precompute_ts = value
                self._EigMultiplier = np.empty((0,0), dtype=self._dtype)
                self._EigValsUsed = np.empty(0, dtype=self._dtype)
                self._EigValsOrig = np.empty(0, dtype=self._dtype)
                self._coef_precomputed = np.empty((0,0), dtype=self._dtype)
            else:
                assert self.precompute_ts_multiplier > 0.
                cy_funs = _wrapper_float if self._use_float else _wrapper_double
                self._EigMultiplier, self._EigValsUsed, self._EigValsOrig = \
                    cy_funs.get_mvnorm_multiplier(self._XtX
                                                    if (self.method == "chol")
                                                    else self._invXtX,
                                                  self.precompute_ts_multiplier,
                                                  self.method != "chol", False)
                if self.n_presampled is not None:
                    self._presample()
                self._precompute_ts = value
        else:
            self._precompute_ts = value

    @property
    def precompute_ts_multiplier(self):
        return float(self._precompute_ts_multiplier)
    
    @precompute_ts_multiplier.setter
    def precompute_ts_multiplier(self, value):
        if (not self._precompute_ts) or (not self.is_fitted_):
            self._precompute_ts_multiplier = value
        elif value != self._precompute_ts_multiplier:
            assert value > 0.
            if not isinstance(value, float):
                value = float(value)
            cy_funs = _wrapper_float if self._use_float else _wrapper_double
            self._EigMultiplier, self._EigValsUsed = \
                cy_funs.mvnorm_from_Eig_different_m(np.empty(0, dtype=self._dtype),
                                                    self._EigMultiplier,
                                                    self._EigValsUsed,
                                                    self._EigValsOrig,
                                                    value, 0, np.random,
                                                    True)
            if self.n_presampled is not None:
                self._presample()
            self._precompute_ts_multiplier = value

    @property
    def n_presampled(self):
        return self._n_presampled
    
    @n_presampled.setter
    def n_presampled(self, value):
        if value != self.n_presampled:
            if value is None:
                self._coef_precomputed = np.empty((0,0), dtype=self._dtype)
            else:
                if (self.is_fitted_) and (self._precompute_ts):
                    self._presample()
        self.n_presampled = n_presampled

    def _set_rng(self):
        if isinstance(self.rng_presample, np.random.RandomState):
            self.rng_presample = self.rng_presample.randint(0, np.iinfo(np.int32).max)
        if isinstance(self.rng_presample, float):
            self.rng_presample = int(self.rng_presample)
        if self.rng_presample is None:
            self.rng_presample = np.random.Generator(np.random.MT19937())
        elif isinstance(self.rng_presample, int):
            self.rng_presample = np.random.Generator(np.random.MT19937(seed = self.rng_presample))

    def _presample(self):
        self._set_rng()
        cy_funs = _wrapper_float if self._use_float else _wrapper_double
        self._coef_precomputed = \
            cy_funs.mvnorm_from_Eig(self.coef_,
                                    self._EigMultiplier,
                                    self._n_presampled,
                                    self.rng_presample)

    def _process_X_y_w(self, X, y, sample_weight, only_X=False):
        if X.dtype != self._dtype:
            X = X.astype(self._dtype)
        if not issparse(X):
            Xcsr = None
            if X.__class__.__name__ == "DataFrame":
                X = X.to_numpy()
            if (not isinstance(X, np.ndarray)) or (not _is_row_major(X)):
                X = np.ascontiguousarray(X)
            if len(X.shape) != 2:
                raise ValueError("'X' must be 2-dimensional")
            X = _copy_if_subview(X)
        else:
            if not isspmatrix_csr(X):
                warnings.warn("Sparse matrices only supported in CSR format. Input will be converted.")
                X = csr_matrix(X)
            Xcsr = X
            X = np.empty((0,0), dtype=self._dtype)
            if len(Xcsr.shape) != 2:
                raise ValueError("'X' must be 2-dimensional")
        if only_X:
            return X, Xcsr

        if issparse(y):
            raise ValueError("Sparse 'y' not supported.")
        if issparse(sample_weight):
            raise ValueError("Sparse 'sample_weight' not supported.")

        y = np.array(y).reshape(-1)
        if sample_weight is None:
            w = np.empty(0, dtype=self._dtype)
        else:
            w = sample_weight

        y = _copy_if_subview_1d(y)
        w = _copy_if_subview_1d(w)

        if Xcsr is None:
            assert X.shape[0] == y.shape[0]
        else:
            if Xcsr.shape[0] != y.shape[0]:
                raise ValueError("'X' and 'y' must have the same number of rows.")

        if sample_weight is not None:
            if X.shape[0] != w.shape[0]:
                raise ValueError("'sample_weight' must have the same number of rows as 'X' and 'y'.")

        if y.dtype != self._dtype:
            y = y.astype(self._dtype)
        if w.dtype != self._dtype:
            w = w.astype(self._dtype)

        return X, Xcsr, y, w

    def fit(self, X, y, sample_weight=None):
        """
        Fit model to data

        Note
        ----
        Calling 'fit' will reset whatever previous data was there. For fitting
        the model incrementally to new data, use 'partial_fit' instead.

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.
        y : array-like(m)
            The target variable.
        sample_weight : None or array-like(m)
            Observation weights for each row.

        Returns
        -------
        self

        """
        assert self.method in ["chol", "sm"]
        assert self.precompute_ts_multiplier > 0

        self._n = X.shape[1]
        X, Xcsr, y, w = self._process_X_y_w(X, y, sample_weight)

        cy_funs = _wrapper_float if self._use_float else _wrapper_double
        if (self.method == "chol") or (X.shape[0] >= X.shape[1]):
            self._XtX, self._invXtX, self._XtY, self.coef_ = \
                cy_funs.fit_model_noinv(
                    X, y, w, Xcsr,
                    add_bias=self.fit_intercept,
                    lam = self.lambda_,
                    calc_inv=(self.calc_inv) or (self.method != "chol")
                )
            if self._precompute_ts:
                self._EigMultiplier, self._EigValsUsed, self._EigValsOrig = \
                    cy_funs.get_mvnorm_multiplier(self._XtX,
                                                  self.precompute_ts_multiplier,
                                                  False, self.method != "chol")
                if self.n_presampled is not None:
                    self._presample()
        else:
            self._invXtX, self._XtY, self.coef_ = \
                cy_funs.fit_model_inv(
                    X, y, w, Xcsr,
                    add_bias=self.fit_intercept,
                    lam = self.lambda_
                )
            if self._precompute_ts:
                self._EigMultiplier, self._EigValsUsed, self._EigValsOrig = \
                    cy_funs.get_mvnorm_multiplier(self._invXtX,
                                                  self.precompute_ts_multiplier,
                                                  True, False)
                if self.n_presampled is not None:
                    self._presample()

        if self.method != "chol":
            self._bufferX = np.empty(self._n + int(self.fit_intercept), dtype=self._dtype)

        self.is_fitted_ = True
        return self


    def partial_fit(self, X, y, sample_weight=None, *args, **kwargs):
        """
        Fit model incrementally to new data

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.
        y : array-like(m)
            The target variable.
        sample_weight : None or array-like(m)
            Observation weights for each row.

        Returns
        -------
        self

        """
        if not self.is_fitted_:
            return self.fit(X, y, sample_weight)

        if X.shape[1] != self._n:
            raise ValueError("Number of columns in X doesn't match with previous.")
        X, Xcsr, y, w = self._process_X_y_w(X, y, sample_weight)

        cy_funs = _wrapper_float if self._use_float else _wrapper_double
        if self.method == "chol":
            if self._invXtX.shape[0] == 0:
                self._invXtX = np.empty((self._XtX.shape[0], self._XtX.shape[1]),
                                        dtype=ctypes.c_float if self._use_float else ctypes.c_double)
            cy_funs.update_running_noinv(
                self._XtX, self._XtY, self._invXtX, self.coef_,
                X, y, w, Xcsr,
                add_bias=self.fit_intercept,
                calc_inv=self.calc_inv
            )
            if self._precompute_ts:
                self._EigMultiplier, self._EigValsUsed, self._EigValsOrig = \
                    cy_funs.get_mvnorm_multiplier(self._XtX,
                                                  self.precompute_ts_multiplier,
                                                  False, False)
                if self.n_presampled is not None:
                    self._presample()
        else:
            if not self._bufferX.shape[0]:
                self._bufferX = np.empty(self._n + int(self.fit_intercept), dtype=self._dtype)
            cy_funs.update_running_inv(
                self._invXtX,
                self._XtY,
                self.coef_,
                self._bufferX,
                X, y, w, Xcsr,
                add_bias=self.fit_intercept,
            )
            if self._precompute_ts:
                self._EigMultiplier, self._EigValsUsed, self._EigValsOrig = \
                    cy_funs.get_mvnorm_multiplier(self._invXtX,
                                                  self.precompute_ts_multiplier,
                                                  True, False)
                if self.n_presampled is not None:
                    self._presample()
        return self

    def predict(self, X):
        """
        Make predictions on new data

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.

        Returns
        -------
        y_hat : array(m)
            The predicted values given 'X'.
        """
        assert self.is_fitted_

        pred = X.dot(self.coef_[:self._n])
        if self.fit_intercept:
            pred[:] += self.coef_[self._n]
        return pred

    def _multiply_x_A_x(self, X):
        X, Xcsr = self._process_X_y_w(X, None, None, only_X=True)
        cy_funs = _wrapper_float if self._use_float else _wrapper_double
        return cy_funs.x_A_x_batch(X, self._invXtX, Xcsr, self.fit_intercept)

    def predict_ucb(self, X, alpha=1.0, add_unfit_noise=False, random_state=None):
        """
        Make an upper-bound prediction on new data

        Make a prediction on new data with an upper bound given by the LinUCB
        formula (be aware that it's not probabilistic like a regular CI).

        Note
        ----
        If using this method, it's recommended to center the 'X' data passed
        to 'fit' and 'partial_fit'. If not centered, it's recommendable to
        lower the ``alpha`` value.

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.
        alpha : float > 0 or array(m, ) > 0
            The multiplier for the width of the bound. Can also pass an array
            with different values for each row.
        add_unfit_noise : bool
            When making predictions with an unfit model (in this case they are
            given by empty zero matrices except for the inverse diagonal matrix
            based on the regularization parameter), whether to add a very small
            amount of random noise ~ Uniform(0, 10^-12) to it. This is useful in
            order to break ties at random when using multiple models.
        random_state : None, np.random.Generator, or np.random.RandomState
            A NumPy 'Generator' or 'RandomState' object instance to use for generating
            random numbers. If passing 'None', will use NumPy's random
            module directly (which can be made reproducible through
            ``np.random.seed``). Only used when passing ``add_unfit_noise=True``
            and calling this method on a model that has not been fit to data.

        Returns
        -------
        y_hat : array(m)
            The predicted upper bound on 'y' given 'X' and ``alpha``.

        References
        ----------
        .. [1] Chu, Wei, et al. "Contextual bandits with linear payoff functions."
               Proceedings of the Fourteenth International Conference on Artificial Intelligence and Statistics. 2011.
        """
        if not isinstance(alpha, np.ndarray):
            assert alpha > 0.
            if isinstance(alpha, int):
                alpha = float(alpha)
            assert isinstance(alpha, float)
        else:
            assert alpha.shape[0] == X.shape[0]

        if (self.method == "chol") and (not self.calc_inv):
            raise ValueError("Not available when using 'method=\"chol\"' and 'calc_inv=False'.")

        if not self.is_fitted_:
            pred = alpha * np.sqrt(np.einsum("ij,ij->i", X, X) / self.lambda_)
            if add_unfit_noise:
                if random_state is None:
                    noise = np.random.uniform(low=0., high=1e-12, size=X.shape[0])
                else:
                    noise = random_state.uniform(low=0., high=1e-12, size=X.shape[0])
                pred[:] += noise
            return pred

        pred = self.predict(X)
        ci = self._multiply_x_A_x(X)
        pred[:] += alpha * np.sqrt(ci)
        return pred

    def predict_thompson(self, X, v_sq=1.0, sample_unique=False, random_state=None):
        """
        Make a guess prediction on new data

        Make a prediction on new data with coefficients sampled from their
        estimated distribution.

        Note
        ----
        If using this method, it's recommended to center the 'X' data passed
        to 'fit' and 'partial_fit'. If not centered, it's recommendable to
        lower the ``v_sq`` value.

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.
        v_sq : float > 0
            The multiplier for the covariance matrix. Larger values lead to
            more variable results.
        sample_unique : bool
            Whether to sample different coefficients each time a prediction is to
            be made. If passing 'False', when calling 'predict', it will sample
            the same coefficients for all the observations in the same call to
            'predict', whereas if passing 'True', will use a different set of
            coefficients for each observations. Passing 'False' leads to an
            approach which is theoretically wrong, but as sampling coefficients
            can be very slow, using 'False' can provide a reasonable speed up
            without much of a performance penalty.
        random_state : None, np.random.Generator, or np.random.RandomState
            A NumPy 'Generator' or 'RandomState' object instance to use for generating
            random numbers. If passing 'None', will use NumPy's random
            module directly (which can be made reproducible through
            ``np.random.seed``).

        Returns
        -------
        y_hat : array(m)
            The predicted guess on 'y' given 'X' and ``v_sq``.

        References
        ----------
        .. [1] Agrawal, Shipra, and Navin Goyal.
               "Thompson sampling for contextual bandits with linear payoffs."
               International Conference on Machine Learning. 2013.
        """
        assert self.is_fitted_
        if X.shape[1] != (self.coef_.shape[0]-int(self.fit_intercept)):
            raise ValueError("'X' has wrong number of columns.")
        assert v_sq > 0.
        if isinstance(v_sq, int):
            v_sq = float(v_sq)
        assert isinstance(v_sq, float)

        cy_funs = _wrapper_float if self._use_float else _wrapper_double
        random_state = np.random if random_state is None else random_state

        if not self._precompute_ts:
            tol = 1e-15

            if self.method != "chol":
                if np.linalg.det(self._invXtX) >= tol:
                    inv_cov = self._invXtX
                else:
                    inv_cov = self._invXtX.copy()
                    n = inv_cov.shape[1]
                    for i in range(10):
                        inv_cov[np.arange(n), np.arange(n)] += 1e-1
                        if np.linalg.det(inv_cov) >= tol:
                            break
                    np.nan_to_num(inv_cov, copy=False)
            else:
                if np.linalg.det(self._XtX) >= tol:
                    cov = self._XtX
                else:
                    cov = self._XtX.copy()
                    n = cov.shape[1]
                    for i in range(10):
                        cov[np.arange(n), np.arange(n)] += 1e-1
                        if np.linalg.det(cov) >= tol:
                            break
                    np.nan_to_num(cov, copy=False)



        if (self.n_presampled is not None) and (np.abs(v_sq - self.precompute_ts_multiplier) <= 1e-3):
            self._set_rng()
            ix_take = self.rng_presample.integers(self.n_presampled, size=X.shape[0], replace=True)
            coef = self._coef_precomputed[ix_take]
            if not issparse(X):
                pred = np.einsum("ij,ij->i", X, coef[:, :X.shape[1]])
            else:
                pred = X.multiply(coef[:, :X.shape[1]]).sum(axis=1)
            if self.fit_intercept:
                pred[:] += coef[:, -1]
        elif sample_unique:
            if self._precompute_ts:
                if np.abs(v_sq - self.precompute_ts_multiplier) <= 1e-3:
                    coef = cy_funs.mvnorm_from_Eig(self.coef_,
                                                   self._EigMultiplier,
                                                   X.shape[0],
                                                   random_state)
                else:
                    coef = cy_funs.mvnorm_from_Eig_different_m(self.coef_,
                                                               self._EigMultiplier,
                                                               self._EigValsUsed,
                                                               self._EigValsOrig,
                                                               v_sq,
                                                               X.shape[0],
                                                               random_state,
                                                               False)
            elif self.method == "chol":
                coef = random_state.multivariate_normal(mean=self.coef_,
                                                        cov=v_sq * cov,
                                                        size=X.shape[0],
                                                        method="cholesky")
            else:
                coef = cy_funs.mvnorm_from_invcov(self.coef_,
                                                  (1./v_sq) * inv_cov,
                                                  size=X.shape[0],
                                                  rng=random_state)
            if not issparse(X):
                pred = np.einsum("ij,ij->i", X, coef[:, :X.shape[1]])
            else:
                pred = X.multiply(coef[:, :X.shape[1]]).sum(axis=1)
            if self.fit_intercept:
                pred[:] += coef[:, -1]
        else:
            if self._precompute_ts:
                if np.abs(v_sq - self.precompute_ts_multiplier) <= 1e-3:
                    coef = cy_funs.mvnorm_from_Eig(self.coef_,
                                                   self._EigMultiplier,
                                                   1,
                                                   random_state)
                else:
                    coef = cy_funs.mvnorm_from_Eig_different_m(self.coef_,
                                                               self._EigMultiplier,
                                                               self._EigValsUsed,
                                                               self._EigValsOrig,
                                                               v_sq,
                                                               1,
                                                               random_state,
                                                               False)
            elif self.method == "chol":
                coef = random_state.multivariate_normal(mean=self.coef_,
                                                        cov=v_sq * cov,
                                                        method="cholesky")
            else:
                coef = cy_funs.mvnorm_from_invcov(self.coef_,
                                                  (1./v_sq) * inv_cov,
                                                  size=1,
                                                  rng=random_state)
            coef = coef.reshape(-1)
            pred = X.dot(coef[:X.shape[1]])
            if self.fit_intercept:
                pred[:] += coef[-1]

        return pred

class ElasticNet(BaseEstimator):
    """
    ElasticNet Regression

    ElasticNet regression (with penalization on the l1 and l2 norms of
    the coefficients),  which keeps track of the aggregated data
    needed to obtain the optimal coefficients in such a way that calling 'partial_fit'
    multiple times would be equivalent to a single call to 'fit' with all the data.
    This is an exact method rather than a stochastic optimization procedure.
    
    Note
    ----
    This ElasticNet regression is fit through a reduction to non-negative
    least squares with twice the number of variables, which is in turn solved
    through a coordinate descent procedure. This is typically slower than the
    lasso paths used in GLMNet and SciKit-Learn, and scales much worse with
    the number of features/columns, but allows for faster incremental updates
    through 'partial_fit', which will give the same result as calls to fit.

    Note
    ----
    This model will not standardize the input data in any way.

    Note
    ----
    By default, will set the l1 and l2 regularization in the same way as
    GLMNet and SciKit-Learn - that is, the regularizations increase along
    with the number of rows in the data, which means they will be different
    after each call to 'fit' or 'partial_fit'. It is nevertheless possible
    to specify the l1 and l2 regularization directly, and both will remain
    constant that way, but be careful about the choice for such hyperparameters.

    Note
    ----
    Doing regression this way requires both memory and computation time
    which scale quadratically with the number of columns/features/variables. As
    such, the class will by default use C 'float' types (typically ``np.float32``)
    instead of C 'double' (``np.float64``), in order to save memory.

    Parameters
    ----------
    alpha : float
        Strenght of the regularization.
    l1_ratio : float [0,1]
        Proportion of the regularization that will be applied to the l1 norm of
        the coefficients (remainder will be applied to the l2 norm). Must be a
        number between zero and one. If passing ``l1_ratio=0``, it's recommended
        instead to use the ``LinearRegression`` class which uses more efficient
        procedures.

        Using higher l1 regularization is more likely to result in some of the
        obtained coefficients being exactly zero, which is oftentimes desirable.
    fit_intercept : bool
        Whether to add an intercept term to the formula. If passing 'True', it will
        be the last entry in the coefficients.
    l1 : None or float
        Strength of the l1 regularization. If passing it, will bypass the values
        set through ``alpha`` and ``l1_ratio``, and will remain constant inbetween
        calls to ``fit`` and ``partial_fit``. If passing this, should also pass
        ``l2`` or otherwise will assume that it is zero.
    l2 : None or float
        Strength of the l2 regularization. If passing it, will bypass the values
        set through ``alpha`` and ``l1_ratio``, and will remain constant inbetween
        calls to ``fit`` and ``partial_fit``. If passing this, should also pass
        ``l1`` or otherwise will assume that it is zero.
    use_float : bool
        Whether to use C 'float' type for the required matrices. If passing 'False',
        will use C 'double'. Be aware that memory usage for this model can grow
        very large. Can be changed after initialization.

    Attributes
    ----------
    coef_ : array(n) or array(n+1)
        The obtained coefficients. If passing 'fit_intercept=True', the intercept
        will be at the last entry.

    References
    ----------
    .. [1] Franc, Vojtech, Vaclav Hlavac, and Mirko Navara. "Sequential coordinate-wise algorithm for the non-negative least squares problem." International Conference on Computer Analysis of Images and Patterns. Springer, Berlin, Heidelberg, 2005.
    """
    def __init__(self, alpha=0.1, l1_ratio=0.5, fit_intercept=True,
                 l1=None, l2=None, use_float=True):
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.fit_intercept = fit_intercept
        self.l1 = l1
        self.l2 = l2
        self._use_float = bool(use_float)

        self._set_dtype(force_cast=False)
        self.coef_ = np.empty(0, dtype=self._dtype)
        self._XtX = np.empty((0,0), dtype=self._dtype)
        self._XtY = np.empty(0, dtype=self._dtype)
        self._buffer_XtY2 = np.empty(0, dtype=self._dtype)
        self._prev_coef = np.empty(0, dtype=self._dtype)
        self._m = 0
        self._n = 0
        self._l1_curr = 0.
        self._l2_curr = 0.

        self.is_fitted_ = False

    def _set_dtype(self, force_cast=False):
        self._dtype = ctypes.c_float if self._use_float else ctypes.c_double

        if force_cast:
            if self.coef_.dtype != self._dtype:
                self.coef_ = self.coef_.astype(self._dtype)
            if self._XtX.dtype != self._dtype:
                self._XtX = self._XtX.astype(self._dtype)
            if self._XtY.dtype != self._dtype:
                self._XtY = self._XtY.astype(self._dtype)
            if self._buffer_XtY2.dtype != self._dtype:
                self._buffer_XtY2 = self._buffer_XtY2.astype(self._dtype)
            if self._prev_coef.dtype != self._dtype:
                self._prev_coef = self._prev_coef.astype(self._dtype)

    @property
    def use_float(self):
        return self._use_float
    @use_float.setter
    def use_float(self, value):
        self._use_float = use_float
        self._set_dtype(force_cast=True)

    def _validate_inputs(self):
        if isinstance(self.alpha, int):
            self.alpha = float(self.alpha)
        if isinstance(self.l1_ratio, int):
            self.l1_ratio = float(self.l1_ratio)
        assert self.alpha >= 0.
        assert self.l1_ratio >= 0.
        assert self.l1_ratio <= 1.

        self.fit_intercept = bool(self.fit_intercept)

        if self.l1 is not None:
            if isinstance(self.l1, int):
                self.l1 = float(self.l1)
            assert self.l1 >= 0.
            if self.l2 is None:
                self.l2 = 0.
        if self.l2 is not None:
            if isinstance(self.l2, int):
                self.l2 = float(self.l2)
            assert self.l2 >= 0.
            if self.l1 is None:
                self.l1 = 0.

    def _calc_l1_l2(self, m):
        if self.l1 is not None:
            return self.l1, self.l2
        div = 1. / m
        l1 = self.alpha * self.l1_ratio
        l2 = self.alpha * (1. - self.l1_ratio)
        l1 /= div
        l2 /= div
        return l1, l2

    def fit(self, X, y, sample_weight=None):
        """
        Fit model to data

        Note
        ----
        Calling 'fit' will reset whatever previous data was there. For fitting
        the model incrementally to new data, use 'partial_fit' instead.

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.
        y : array-like(m)
            The target variable.
        sample_weight : None or array-like(m)
            Observation weights for each row.

        Returns
        -------
        self

        """
        self._validate_inputs()
        X, Xcsr, y, w = LinearRegression._process_X_y_w(self, X, y, sample_weight)
        self._m = X.shape[0]
        self._n = X.shape[1]
        if min(self._m, self._n) == 0:
            raise ValueError("Cannot fit model to empty inputs.")
        l1, l2 = self._calc_l1_l2(self._m)

        n_plus_b = int(self._n + self.fit_intercept)
        self._XtX = np.zeros((n_plus_b, n_plus_b), dtype=self._dtype)
        self._XtY = np.zeros(n_plus_b, dtype=self._dtype)
        self._buffer_XtY2 = np.empty(2*n_plus_b, dtype=self._dtype)
        self._prev_coef = np.zeros(2*n_plus_b, dtype=self._dtype)
        self.coef_ = np.empty(n_plus_b, dtype=self._dtype)

        cy_funs = _wrapper_float if self._use_float else _wrapper_double
        cy_funs.update_matrices_noinv(
            X,
            y,
            w,
            self._XtX,
            self._XtY,
            Xcsr,
            bool(self.fit_intercept),
            True
        )
        if (Xcsr is None) and np.isfortran(X):
            self._XtX[:,:] = self._XtX.T
            if np.isfortran(self._XtX):
                self._XtX = np.ascontiguousarray(self._XtX)
        ### Do not regularize the intercept
        cy_funs.add_to_diag(self._XtX, self.fit_intercept, l2)
        cy_funs.fill_lower_triangle(self._XtX)
        self._buffer_XtY2[:n_plus_b] =  self._XtY - l1
        self._buffer_XtY2[n_plus_b:] = -self._XtY - l1
        ### Do not regularize the intercept
        if self.fit_intercept:
            self._buffer_XtY2[self._n] =  self._XtY[-1]
            self._buffer_XtY2[-1]      = -self._XtY[-1]
        cy_funs.solve_elasticnet(
            self._XtX,
            self._buffer_XtY2,
            self._prev_coef,
            self.coef_
        )

        self._l1_curr = l1
        self._l2_curr = l2
        self.is_fitted_ = True
        return self

    def partial_fit(self, X, y, sample_weight=None, *args, **kwargs):
        """
        Fit model incrementally to new data

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.
        y : array-like(m)
            The target variable.
        sample_weight : None or array-like(m)
            Observation weights for each row.

        Returns
        -------
        self

        """
        if not self.is_fitted_:
            return self.fit(X, y, sample_weight)

        if X.shape[1] != self._n:
            raise ValueError("Number of columns in X doesn't match with previous.")
        m_add = X.shape[0]
        if m_add == 0:
            return self
        self._validate_inputs()
        X, Xcsr, y, w = LinearRegression._process_X_y_w(self, X, y, sample_weight)

        ### Note: the regularization changes at each step if imitating elasticnet
        l1, l2_target = self._calc_l1_l2(self._m + m_add)
        l2_add = l2_target - self._l2_curr
        n_plus_b = int(self._n + self.fit_intercept)

        cy_funs = _wrapper_float if self._use_float else _wrapper_double
        cy_funs.update_matrices_noinv(
            X,
            y,
            w,
            self._XtX,
            self._XtY,
            Xcsr,
            bool(self.fit_intercept),
            False
        )
        if (Xcsr is None) and np.isfortran(X):
            self._XtX[:,:] = self._XtX.T
            if np.isfortran(self._XtX):
                self._XtX = np.ascontiguousarray(self._XtX)
        if (np.abs(l2_add) > 1e-10):
            cy_funs.add_to_diag(self._XtX, self.fit_intercept, l2_add)
        cy_funs.fill_lower_triangle(self._XtX)
        self._buffer_XtY2[:n_plus_b] =  self._XtY - l1
        self._buffer_XtY2[n_plus_b:] = -self._XtY - l1
        ### Do not regularize the intercept
        if self.fit_intercept:
            self._buffer_XtY2[self._n] =  self._XtY[-1]
            self._buffer_XtY2[-1]      = -self._XtY[-1]
        ### TODO: for some reason, it gets a very bad result if the previous
        ### coefficients are reused, which should not be the case
        self._prev_coef[:] = 0.
        cy_funs.solve_elasticnet(
            self._XtX,
            self._buffer_XtY2,
            self._prev_coef,
            self.coef_
        )

        self._m += m_add
        self._l1_curr = l1
        self._l2_curr = l2_target
        return self

    def predict(self, X):
        """
        Make predictions on new data

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.

        Returns
        -------
        y_hat : array(m)
            The predicted values given 'X'.
        """
        return LinearRegression.predict(self, X)



import numpy as np
cimport numpy as np
import ctypes
from libc.string cimport memcpy, memset
from libc.math cimport isnan, isfinite
from cython cimport boundscheck, nonecheck, wraparound
from libc.stdio cimport printf

ctypedef enum CBLAS_ORDER:
    CblasRowMajor = 101
    CblasColMajor = 102

ctypedef CBLAS_ORDER CBLAS_LAYOUT

ctypedef enum CBLAS_TRANSPOSE:
    CblasNoTrans=111
    CblasTrans=112
    CblasConjTrans=113
    CblasConjNoTrans=114

ctypedef enum CBLAS_UPLO:
    CblasUpper=121
    CblasLower=122

ctypedef enum CBLAS_DIAG:
    CblasNonUnit=131
    CblasUnit=132

ctypedef enum CBLAS_SIDE:
    CblasLeft=141
    CblasRight=142

cdef void cblas_tsyrk(
        CBLAS_ORDER Order, CBLAS_UPLO Uplo, CBLAS_TRANSPOSE Trans,
        int N, int K, real_t alpha, real_t *A, int lda, real_t beta, real_t *C, int ldc
    ) nogil:
    cdef char uplo
    cdef char trans
    if (Order == CblasColMajor):
        if (Uplo == CblasUpper):
            uplo = 'U'
        else:
            uplo = 'L'

        if (Trans == CblasTrans):
            trans = 'T'
        elif (Trans == CblasConjTrans):
            trans = 'C'
        else:
            trans = 'N'

    else:
        if (Uplo == CblasUpper):
            uplo = 'L'
        else:
            uplo = 'U'

        if (Trans == CblasTrans):
            trans = 'N'
        elif (Trans == CblasConjTrans):
            trans = 'N'
        else:
            trans = 'T'

    tsyrk(&uplo, &trans, &N, &K, &alpha, A, &lda, &beta, C, &ldc)

cdef void cblas_tgemm(
        CBLAS_ORDER Order, CBLAS_TRANSPOSE TransA, CBLAS_TRANSPOSE TransB, int M, int N, int K,
        real_t alpha, real_t *A, int lda, real_t *B, int ldb, real_t beta, real_t *C, int ldc
    ) nogil:
    cdef char transA
    cdef char transB

    if (Order == CblasColMajor):
        if (TransA == CblasTrans):
            transA = 'T'
        elif (TransA == CblasConjTrans):
            transA = 'C'
        else:
            transA = 'N'

        if (TransB == CblasTrans):
            transB = 'T'
        elif (TransB == CblasConjTrans):
            transB = 'C'
        else:
            transB = 'N'

        tgemm(&transA, &transB, &M, &N, &K, &alpha, A, &lda, B, &ldb, &beta, C, &ldc)

    else:
        if (TransA == CblasTrans):
            transB = 'T'
        elif (TransA == CblasConjTrans):
            transB = 'C'
        else:
            transB = 'N'

        if (TransB == CblasTrans):
            transA = 'T';
        elif (TransB == CblasConjTrans):
            transA = 'C'
        else:
            transA = 'N'

        tgemm(&transA, &transB, &N, &M, &K, &alpha, B, &ldb, A, &lda, &beta, C, &ldc)

cdef void cblas_tgemv(
        CBLAS_ORDER order,  CBLAS_TRANSPOSE TransA,  int m, int n,
        real_t alpha, real_t  *a, int lda,  real_t  *x, int incx,  real_t beta,  real_t  *y, int incy
    ) nogil:
    cdef char trans
    if (order == CblasColMajor):
        if (TransA == CblasNoTrans):
            trans = 'N';
        elif (TransA == CblasTrans):
            trans = 'T'
        else:
            trans = 'C'
        tgemv(&trans, &m, &n, &alpha, a, &lda, x, &incx, &beta, y, &incy)

    else:
        if (TransA == CblasNoTrans):
            trans = 'T'
        elif (TransA == CblasTrans):
            trans = 'N'
        else:
            trans = 'N'

        tgemv(&trans, &n, &m, &alpha, a, &lda, x, &incx, &beta, y, &incy)

cdef void cblas_tsymv(
        CBLAS_LAYOUT layout,
        CBLAS_UPLO Uplo, int N,
        real_t alpha, real_t  *A, int lda,
        real_t  *X, int incX, real_t beta,
        real_t  *Y, int incY
    ) nogil:
    cdef char ul
    if (layout == CblasColMajor):
        if (Uplo == CblasUpper):
            ul = 'U'
        else:
            ul = 'L'
    else:
        if (Uplo == CblasUpper):
            ul = 'L'
        else:
            ul = 'U'
    tsymv(&ul, &N, &alpha, A, &lda, X, &incX, &beta, Y, &incY)

cdef void cblas_tsymm(
        CBLAS_LAYOUT layout, CBLAS_SIDE Side,
        CBLAS_UPLO Uplo, int M, int N,
        real_t alpha, real_t  *A, int lda,
        real_t  *B, int ldb, real_t beta,
        real_t  *C, int ldc
    ) nogil:
    cdef char sd
    cdef char ul

    if (layout == CblasColMajor):
        if (Side == CblasRight):
            sd = 'R'
        else:
            sd = 'L'

        if (Uplo == CblasUpper):
            ul = 'U'
        else:
            ul = 'L'

        tsymm(&sd, &ul, &M, &N, &alpha, A, &lda, B, &ldb, &beta, C, &ldc)

    else:
        if (Side == CblasRight):
            sd = 'L'
        else:
            sd = 'R'

        if (Uplo == CblasUpper):
            ul = 'L'
        else:
            ul = 'U'

        tsymm(&sd, &ul, &N, &M, &alpha, A, &lda, B, &ldb, &beta, C, &ldc)

cdef void cblas_tsyr(
        CBLAS_LAYOUT layout, CBLAS_UPLO Uplo,
        int N, real_t  alpha, real_t  *X,
        int incX, real_t  *A, int lda
    ) nogil:
    cdef char ul

    if (layout == CblasColMajor):
        if (Uplo == CblasLower):
            ul = 'L'
        else:
            ul = 'U'
        tsyr(&ul, &N, &alpha, X, &incX, A, &lda)

    else:
        if (Uplo == CblasLower):
            ul = 'U'
        else:
            ul = 'L'
        tsyr(&ul, &N, &alpha, X, &incX, A, &lda)


cdef void tsymv_dense_sp(
        real_t A[], int n_plusb,
        real_t data[], long indptr[], long indices[], long row,
        real_t outp[], bint add_bias
    ) nogil:
    cdef long i
    cdef long col
    cdef long rowA
    memset(outp, 0, n_plusb*sizeof(real_t))
    for rowA in range(n_plusb):
        for i in range(indptr[row], indptr[row+1]):
            col = indices[i]
            outp[rowA] += A[(col + rowA*n_plusb) if (col >= rowA) else (rowA + col*n_plusb)] * data[i]
    
    cdef int one = 1
    cdef real_t one_real_t = 1.
    if add_bias:
        taxpy(&n_plusb, &one_real_t, A + (n_plusb-1), &n_plusb, outp, &one)

cdef real_t tdot_dense_sp(
        real_t *dense_vec, long row,
        real_t data[], long indptr[], long indices[]
    ) nogil:
    cdef real_t res = 0
    cdef long i
    for i in range(indptr[row], indptr[row+1]):
        res += dense_vec[indices[i]] * data[i]
    return res

cdef real_t* get_ptr_real_t(np.ndarray[real_t, ndim=1] arr):
    return &arr[0]

cdef long* get_ptr_long(np.ndarray[long, ndim=1] arr):
    return &arr[0]

def cast_csr(Xcsr):
    if Xcsr.data.dtype != C_real_t:
        Xcsr.data = Xcsr.data.astype(C_real_t)
    if Xcsr.indptr.dtype != ctypes.c_long:
        Xcsr.indptr = Xcsr.indptr.astype(ctypes.c_long)
    if Xcsr.indices.dtype != ctypes.c_long:
        Xcsr.indices = Xcsr.indices.astype(ctypes.c_long)

def fit_model_noinv(
        np.ndarray[real_t, ndim=2] X,
        np.ndarray[real_t, ndim=1] y,
        np.ndarray[real_t, ndim=1] w,
        Xcsr = None,
        bint add_bias=1,
        real_t lam = 1.,
        bint calc_inv=0
    ):
    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]
    cdef np.ndarray[real_t, ndim=2] XtX = np.empty((n_plusb,n_plusb), dtype=C_real_t)
    cdef np.ndarray[real_t, ndim=1] XtY = np.empty(n_plusb, dtype=C_real_t)

    update_matrices_noinv(
        X, y, w,
        XtX, XtY,
        Xcsr,
        add_bias=add_bias,
        overwrite=1
    )

    if lam != 0.:
        XtX[np.arange(n), np.arange(n)] += lam

    cdef np.ndarray[real_t, ndim=2] XtX_copy = XtX.copy()
    cdef np.ndarray[real_t, ndim=1] XtY_copy = XtY.copy()

    cdef char lo = 'L'
    cdef int one_int = 1
    cdef int ignore = 0

    if not calc_inv:
        tposv(&lo, &n_plusb, &one_int, &XtX[0,0], &n_plusb, &XtY[0], &n_plusb, &ignore)
        return XtX_copy, XtX, XtY_copy, XtY
        ## XtX, invXtX, XtY, coef
        ##      (chol)

    else:
        tposv(&lo, &n_plusb, &one_int, &XtX[0,0], &n_plusb, &XtY[0], &n_plusb, &ignore)
        tpotri(&lo, &n_plusb, &XtX[0,0], &n_plusb, &ignore)
        return XtX_copy, XtX, XtY_copy, XtY
        ## XtX, invXtX, XtY, coef


def fit_model_inv(
        np.ndarray[real_t, ndim=2] X,
        np.ndarray[real_t, ndim=1] y,
        np.ndarray[real_t, ndim=1] w,
        Xcsr = None,
        bint add_bias=1,
        real_t lam = 1.
    ):

    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias

    cdef np.ndarray[real_t, ndim=2] invXtX = np.zeros((n_plusb,n_plusb), dtype=C_real_t)
    cdef np.ndarray[real_t, ndim=1] XtY = np.empty(n_plusb, dtype=C_real_t)

    invXtX[np.arange(n_plusb), np.arange(n_plusb)] = 1. / lam

    if np.isfortran(X):
        X = np.ascontiguousarray(X)
    if X.dtype != C_real_t:
        X = X.astype(C_real_t)

    update_matrices_inv(
        invXtX,
        XtY,
        XtY,
        X, y, w,
        Xcsr = Xcsr,
        add_bias=add_bias,
        overwrite=1
    )

    cdef real_t *ptr_XtY = &XtY[0]
    cdef real_t *ptr_invXtX = &invXtX[0,0]

    cdef np.ndarray[real_t, ndim=1] coefs = np.empty(n_plusb, dtype=C_real_t)
    cblas_tsymv(CblasRowMajor, CblasUpper, n_plusb,
                1., ptr_invXtX, n_plusb, ptr_XtY, 1,
                0., &coefs[0], 1)

    return invXtX, XtY, coefs


def update_running_noinv(
        np.ndarray[real_t, ndim=2] XtX,
        np.ndarray[real_t, ndim=1] XtY,
        np.ndarray[real_t, ndim=2] invXtX,
        np.ndarray[real_t, ndim=1] coef,
        np.ndarray[real_t, ndim=2] X,
        np.ndarray[real_t, ndim=1] y,
        np.ndarray[real_t, ndim=1] w,
        Xcsr = None,
        bint add_bias=1,
        bint calc_inv=0
    ):

    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]

    cdef np.ndarray[real_t, ndim=2] Xw

    cdef char lo = 'L'
    cdef int one_int = 1
    cdef int ignore = 0

    update_matrices_noinv(
        X, y, w,
        XtX, XtY,
        Xcsr = Xcsr,
        add_bias=add_bias,
        overwrite=0
    )

    coef[:] = XtY
    invXtX[:,:] = XtX[:,:]
    tposv(&lo, &n_plusb, &one_int, &invXtX[0,0], &n_plusb, &coef[0], &n_plusb, &ignore)
    if calc_inv:
        tpotri(&lo, &n_plusb, &invXtX[0,0], &n_plusb, &ignore)

def update_running_inv(
        np.ndarray[real_t, ndim=2] invXtX,
        np.ndarray[real_t, ndim=1] XtY,
        np.ndarray[real_t, ndim=1] coefs,
        np.ndarray[real_t, ndim=1] x_vec, ### buffer
        np.ndarray[real_t, ndim=2] X,
        np.ndarray[real_t, ndim=1] y,
        np.ndarray[real_t, ndim=1] w,
        Xcsr = None,
        bint add_bias=1
    ):

    if np.isfortran(X):
        X = np.ascontiguousarray(X)

    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    
    update_matrices_inv(
        invXtX, XtY, x_vec,
        X, y, w,
        Xcsr = Xcsr,
        add_bias=add_bias,
        overwrite=0
    )

    cdef real_t *ptr_XtY = &XtY[0]
    cdef real_t *ptr_invXtX = &invXtX[0,0]
    cblas_tsymv(CblasRowMajor, CblasUpper, n_plusb,
                1., ptr_invXtX, n_plusb, ptr_XtY, 1,
                0., &coefs[0], 1)


def update_matrices_noinv(
        np.ndarray[real_t, ndim=2] X,
        np.ndarray[real_t, ndim=1] y,
        np.ndarray[real_t, ndim=1] w,
        np.ndarray[real_t, ndim=2] XtX,
        np.ndarray[real_t, ndim=1] XtY,
        Xcsr = None,
        bint add_bias=1,
        bint overwrite=0
    ):
    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]

    cdef np.ndarray[real_t, ndim=1] Xsum
    cdef np.ndarray[real_t, ndim=2] Xw

    ### Note: in theory, can use X as either row-major or column-major,
    ### but then would have to keep track of whether XtX if upper or lower triangular,
    ### since it will produce different structures according to X.
    X = np.require(X, dtype=X.dtype, requirements=["F_CONTIGUOUS"])

    cdef CBLAS_ORDER x_ord = CblasRowMajor if not np.isfortran(X) else CblasColMajor
    if Xcsr is None:
        if w.shape[0] == 0:
            cblas_tsyrk(x_ord, CblasUpper if (x_ord==CblasRowMajor) else CblasLower, CblasTrans,
                        n, m,
                        1., &X[0,0], n if (x_ord==CblasRowMajor) else m,
                        0. if overwrite else 1., &XtX[0,0], n_plusb)
        else:
            Xw = X * w.reshape((-1,1))
            if np.isfortran(Xw) != np.isfortran(X):
                X = np.ascontiguousarray(X)
                Xw = np.ascontiguousarray(Xw)
                x_ord = CblasRowMajor
            cblas_tgemm(x_ord, CblasTrans, CblasNoTrans,
                        n, n, m,
                        1., &Xw[0,0], n if (x_ord==CblasRowMajor) else m,
                        &X[0,0], n if (x_ord==CblasRowMajor) else m,
                        0. if overwrite else 1., &XtX[0,0], n_plusb)
    else:
        if w.shape[0] == 0:
            if overwrite:
                XtX[:n,:n] = np.array(Xcsr.T.dot(Xcsr).todense())
            else:
                XtX[:n,:n] += np.array(Xcsr.T.dot(Xcsr).todense())
        else:
            Xw_sp = Xcsr.multiply(w.reshape((-1,1)))
            if overwrite:
                XtX[:n,:n] = np.array(Xw_sp.T.dot(Xcsr).todense())
            else:
                XtX[:n,:n] += np.array(Xw_sp.T.dot(Xcsr).todense())

    cdef int one_int = 1
    cdef real_t one = 1.
    if add_bias:
        if overwrite:
            XtX[n,n] = 0

        if w.shape[0] == 0:
            if Xcsr is None:
                Xsum = X.sum(axis=0)
            else:
                Xsum = np.array(Xcsr.sum(axis=0)).reshape(-1)
            XtX[n,n] += <real_t>m
        else:
            if Xcsr is None:
                Xsum = np.einsum("ij,i->j", X, w)
            else:
                Xsum = np.array(Xcsr
                                .multiply(w.reshape((-1,1)))
                                .sum(axis=1))\
                                .reshape(-1)
            XtX[n,n] += <real_t> (w.sum())
        if overwrite:
            tcopy(&n, &Xsum[0], &one_int, &XtX[0,n], &n_plusb)
        else:
            taxpy(&n, &one, &Xsum[0], &one_int, &XtX[0,n], &n_plusb)

    if XtY.shape[0]:
        if w.shape[0] == 0:
            update_XtY(
                XtY,
                X, y, np.empty(0, dtype=C_real_t),
                Xcsr = Xcsr,
                add_bias=add_bias,
                overwrite=overwrite
            )
        else:
            update_XtY(
                XtY,
                Xw if Xcsr is None else X, y, np.empty(0, dtype=C_real_t),
                Xcsr = Xw_sp if Xcsr is not None else None,
                add_bias=add_bias,
                overwrite=overwrite
            )

def update_matrices_inv(
        np.ndarray[real_t, ndim=2] invXtX,
        np.ndarray[real_t, ndim=1] XtY,
        np.ndarray[real_t, ndim=1] x_vec, ###buffer
        np.ndarray[real_t, ndim=2] X,
        np.ndarray[real_t, ndim=1] y,
        np.ndarray[real_t, ndim=1] w,
        Xcsr = None,
        bint add_bias=1,
        bint overwrite=0
    ):

    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]

    if np.isfortran(X):
        X = np.ascontiguousarray(X)

    cdef real_t *prt_X = &X[0,0] if Xcsr is None else NULL
    cdef real_t *ptr_x_vec = &x_vec[0]
    cdef real_t *ptr_XtX_inv = &invXtX[0,0]
    cdef real_t *ptr_w = NULL
    cdef real_t coef
    cdef int one = 1
    cdef real_t one_real_t = 1.

    if w.shape[0]:
        ptr_w = &w[0]

    cdef real_t *ptr_csr_data = NULL
    cdef long *ptr_csr_indptr = NULL
    cdef long *ptr_csr_indices = NULL

    cdef long i
    if Xcsr is None:
        with nogil, boundscheck(False), nonecheck(False), wraparound(False):
            for i in range(m):
                cblas_tsymv(CblasRowMajor, CblasUpper, n,
                            1., ptr_XtX_inv, n_plusb, prt_X + i*n, 1,
                            0., ptr_x_vec, 1)
                if (add_bias):
                    ptr_x_vec[n] = tdot(&n, ptr_XtX_inv + n, &n_plusb, prt_X + i*n, &one)
                    ptr_x_vec[n] += ptr_XtX_inv[n_plusb*n_plusb - 1]
                    taxpy(&n, &one_real_t, ptr_XtX_inv + n, &n_plusb, ptr_x_vec, &one)

                coef = tdot(&n, ptr_x_vec, &one, prt_X + i*n, &one)
                if add_bias:
                    coef += ptr_x_vec[n]
                coef = -1. / (1. + coef)
                if ptr_w != NULL:
                    coef *= ptr_w[i]

                cblas_tsyr(CblasRowMajor, CblasUpper,
                           n_plusb, coef, ptr_x_vec, 1,
                           ptr_XtX_inv, n_plusb)

    else:
        cast_csr(Xcsr)
        ptr_csr_data = get_ptr_real_t(Xcsr.data)
        ptr_csr_indices = get_ptr_long(Xcsr.indices)
        ptr_csr_indptr = get_ptr_long(Xcsr.indptr)
        with nogil, boundscheck(False), nonecheck(False), wraparound(False):
            for i in range(m):
                tsymv_dense_sp(
                    ptr_XtX_inv, n_plusb,
                    ptr_csr_data, ptr_csr_indptr, ptr_csr_indices, i,
                    ptr_x_vec, add_bias
                )
                coef = tdot_dense_sp(ptr_x_vec, i,
                                     ptr_csr_data, ptr_csr_indptr, ptr_csr_indices)
                if add_bias:
                    coef += ptr_x_vec[n]
                coef = -1. / (1. + coef)
                if ptr_w != NULL:
                    coef *= ptr_w[i]
                cblas_tsyr(CblasRowMajor, CblasUpper,
                           n_plusb, coef, ptr_x_vec, 1,
                           ptr_XtX_inv, n_plusb)

    if ptr_w != NULL:
        y = y * w

    if XtY.shape[0]:
        update_XtY(
            XtY,
            X, y, w,
            Xcsr = Xcsr,
            add_bias=add_bias,
            overwrite=overwrite
        )


def update_XtY(
        np.ndarray[real_t, ndim=1] XtY,
        np.ndarray[real_t, ndim=2] X,
        np.ndarray[real_t, ndim=1] y,
        np.ndarray[real_t, ndim=1] w,
        Xcsr = None,
        bint add_bias=1,
        bint overwrite=0
    ):

    ## TODO: maybe this one should do the updates row-by-row,
    ## since most of the 'y' entries are expected to be zeros
    cdef CBLAS_ORDER x_ord = CblasRowMajor if not np.isfortran(X) else CblasColMajor
    
    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]

    cdef np.ndarray[real_t, ndim=2] Xw

    if Xcsr is None:
        if w.shape[0] == 0:
            cblas_tgemv(x_ord, CblasTrans,
                        m, n,
                        1., &X[0,0], n if (x_ord==CblasRowMajor) else m,
                        &y[0], 1,
                        0. if overwrite else 1., &XtY[0], 1)
        else:
            Xw = X * w.reshape((-1,1))
            if np.isfortran(Xw) != np.isfortran(X):
                X = np.ascontiguousarray(X)
                Xw = np.ascontiguousarray(Xw)
                x_ord = CblasRowMajor
            cblas_tgemv(x_ord, CblasTrans,
                        m, n,
                        1., &Xw[0,0], n if (x_ord==CblasRowMajor) else m,
                        &y[0], 1,
                        0. if overwrite else 1., &XtY[0], 1)
    
    else:
        if w.shape[0] == 0:
            if overwrite:
                XtY[:n] = Xcsr.T.dot(y)
            else:
                XtY[:n] += Xcsr.T.dot(y)
        else:
            Xw_sp = Xcsr.multiply(w.reshape((-1,1)))
            if overwrite:
                XtY[:n] = Xw_sp.T.dot(y)
            else:
                XtY[:n] += Xw_sp.T.dot(y)

    cdef int one_int = 1
    cdef real_t one = 1.
    if add_bias:
        if overwrite:
            XtY[n] = 0

        if w.shape[0] == 0:
            XtY[n] += y.sum()
        else:
            XtY[n] += tdot(&m, &y[0], &one_int, &w[0], &one_int)
 
def x_A_x_batch(
        np.ndarray[real_t, ndim=2] X,
        np.ndarray[real_t, ndim=2] invXtX,
        Xcsr = None,
        bint add_bias=1,
        bint copy_X=0
    ):
    
    if (X.shape[0] == 0) and (Xcsr.shape[0] == 0):
        return None
    
    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]

    cdef np.ndarray[real_t, ndim=2] tempX
    cdef np.ndarray[real_t, ndim=1] outp = np.empty(m, dtype=C_real_t)

    cdef long i, j
    cdef real_t one = 1.
    cdef int one_int = 1

    cdef real_t *ptr_tempX = NULL
    cdef real_t *ptr_X = NULL
    cdef real_t *ptr_invXtX = &invXtX[0,0]
    cdef real_t *ptr_outp = &outp[0]

    if Xcsr is None:
        X = np.require(X, dtype=C_real_t, requirements=["C_CONTIGUOUS", "OWNDATA", "WRITEABLE"])

        tempX = np.zeros((m, n_plusb), dtype=C_real_t)
        ptr_tempX = &tempX[0,0]
        ptr_X = &X[0,0]

        with nogil, boundscheck(False), nonecheck(False), wraparound(False):
            cblas_tsymm(CblasRowMajor, CblasRight, CblasUpper,
                        m, n,
                        1., ptr_invXtX, n_plusb,
                        ptr_X, n,
                        0., ptr_tempX, n_plusb)
            ## TODO: this is parallelizable, but doesn't get any improvement beyond 2 threads
            for i in range(m):
                ptr_outp[i] = tdot(&n, ptr_tempX + i*n_plusb, &one_int, ptr_X + i*n, &one_int)

            if add_bias:
                cblas_tgemv(CblasRowMajor, CblasNoTrans,
                            m, n,
                            2., ptr_X, n, ptr_invXtX + n, n_plusb,
                            1., ptr_outp, 1)
                for i in range(m):
                    ptr_outp[i] += ptr_invXtX[n_plusb*n_plusb - 1]

    else:
        if Xcsr.dtype != C_real_t:
            Xcsr = Xcsr.astype(C_real_t)

        with nogil, boundscheck(False), nonecheck(False), wraparound(False):
            for i in range(n_plusb):
                for j in range(i):
                    ptr_invXtX[j + i*n_plusb] = ptr_invXtX[i + j*n_plusb]

        tempXcsr = Xcsr.dot(invXtX[:n,:n])
        outp[:] = Xcsr.multiply(tempXcsr).sum(axis=1).reshape(-1)

        if add_bias:
            outp[:] += 2. * Xcsr.dot(invXtX[n,:n])
            outp[:] += invXtX[n,n]

    return outp

def get_matrix_inv(
        np.ndarray[real_t, ndim=2] X,
        np.ndarray[real_t, ndim=2] invX
    ):
    if X.shape[0] == 0:
        return None
    cdef real_t *ptr_X = &X[0,0]
    cdef real_t *ptr_invX = &invX[0,0]
    cdef int n = X.shape[0]
    cdef int ignore = 0
    cdef char lo = 'L'
    with nogil:
        memcpy(ptr_invX, ptr_X, <size_t>(X.shape[0]*X.shape[1])*sizeof(real_t))
        tpotrf(&lo, &n, ptr_invX, &n, &ignore)
        tpotri(&lo, &n, ptr_invX, &n, &ignore)

### https://stats.stackexchange.com/questions/175271/can-i-get-a-cholesky-decomposition-from-the-inverse-of-a-matrix
### Sampling from multivariate normal N(mu, Sigma):
### - Sample Y[n,m] ~ N(0,1)
### - Convert to desired distribution by transforming:
###      A = EigVec(Sigma) * diag(sqrt(EigVal(Sigma)))
###      z = mu + Y * A
#### (If using the inverse of Sigma, the eigen vectors are the same,
####  while the eigen values are 1 divided over the eigen values of the inverse)
def get_mvnorm_multiplier(
        np.ndarray[real_t, ndim=2] Sigma,
        real_t multiplier,
        bint is_inverse,
        bint overwrite_sigma
    ):

    Sigma = np.require(Sigma, dtype=C_real_t, requirements=["C_CONTIGUOUS", "OWNDATA"])
    cdef int m = Sigma.shape[0]
    cdef np.ndarray[real_t, ndim=2] eig_Vecs = np.empty((0,0), dtype=C_real_t)
    if overwrite_sigma:
        eig_Vecs = Sigma
    else:
        eig_Vecs = Sigma.copy()
    cdef np.ndarray[real_t, ndim=1] eig_vals = np.empty(m, dtype=C_real_t)
    cdef real_t *ptr_Ev = &eig_Vecs[0,0]
    cdef real_t *ptr_ev = &eig_vals[0]

    cdef char lo = 'L'
    cdef int one = 1
    cdef int minus_one = -1
    cdef char V = 'V'
    cdef int ignore
    cdef real_t size_buffer

    tsyev(&V, &lo, &m, ptr_Ev, &m, ptr_ev, &size_buffer, &minus_one, &ignore)
    cdef int size_buffer_as_int = int(size_buffer)
    cdef np.ndarray[real_t, ndim=1] buffer_arr = np.empty(size_buffer_as_int, dtype=C_real_t)
    tsyev(&V, &lo, &m, ptr_Ev, &m, ptr_ev, &buffer_arr[0], &size_buffer_as_int, &ignore)

    ### Note: the obtained eigen values are transposed at this point as they were passed
    ### assuming column-major (input was a symmetric matrix so it doesn't matter).
    ### If doing the rest of this part in C, would have to use tscal with incX given by
    ### the number of columns

    ### Note2: If multiplying the input matrix by a constant, the eigen vectors
    ### remain the same, while the eigen values get multiplied by that same constant
    if is_inverse:
        ix_replace = eig_vals <= 1e-10
        eig_vals[ix_replace] = 1.
        eig_vals[:] =  1. / eig_vals
        eig_vals[ix_replace] = 0.
    eig_vals[:] = eig_vals.clip(min=0.)
    eig_vals_orig = eig_vals.copy()
    eig_vals[:] =  np.sqrt(multiplier * eig_vals)

    return np.einsum("ij,i->ij", eig_Vecs, eig_vals), eig_vals, eig_vals_orig


def mvnorm_from_Eig(
        np.ndarray[real_t, ndim=1] mu,
        np.ndarray[real_t, ndim=2] EigMultiplier,
        size_t n,
        rng
    ):
    cdef size_t m = EigMultiplier.shape[0]
    cdef np.ndarray[real_t, ndim=2] X = np.empty((0,0), dtype=C_real_t)
    if isinstance(rng, np.random.Generator):
        X = rng.standard_normal(size=(n, m), dtype=C_real_t)
    else:
        X = rng.normal(size=(n, m), dtype=C_real_t).astype(C_real_t)
    return mu + X.dot(EigMultiplier)


def mvnorm_from_Eig_different_m(
        np.ndarray[real_t, ndim=1] mu,
        np.ndarray[real_t, ndim=2] EigMultiplier,
        np.ndarray[real_t, ndim=1] eig_vals_multiplied,
        np.ndarray[real_t, ndim=1] eig_vals_orig,
        real_t new_m,
        size_t n,
        rng,
        bint return_multiplier
    ):
    new_EigMultiplier = EigMultiplier / eig_vals_multiplied.reshape((-1,1))
    new_EigMultiplier = np.einsum("ij,i->ij",
                                  new_EigMultiplier,
                                  np.sqrt(new_m * eig_vals_orig))
    if return_multiplier:
        return new_EigMultiplier, np.sqrt(new_m * eig_vals_orig)
    else:
        return mvnorm_from_Eig(mu, new_EigMultiplier, n, rng)

## https://stats.stackexchange.com/questions/175271/can-i-get-a-cholesky-decomposition-from-the-inverse-of-a-matrix
def mvnorm_from_invcov(
        np.ndarray[real_t, ndim=1] mu,
        np.ndarray[real_t, ndim=2] invSigma,
        size_t size,
        rng
    ):
    
    ### First calculate eigen values and eigen vectors
    cdef int m = invSigma.shape[0]
    cdef np.ndarray[real_t, ndim=2] eig_Vecs = invSigma.copy()
    cdef np.ndarray[real_t, ndim=1] eig_vals = np.empty(m, dtype=C_real_t)
    cdef real_t *ptr_Ev = &eig_Vecs[0,0]
    cdef real_t *ptr_ev = &eig_vals[0]

    cdef char lo = 'L'
    cdef int one = 1
    cdef int minus_one = -1
    cdef char V = 'V'
    cdef int ignore
    cdef real_t size_buffer

    tsyev(&V, &lo, &m, ptr_Ev, &m, ptr_ev, &size_buffer, &minus_one, &ignore)
    cdef int size_buffer_as_int = int(size_buffer)
    cdef np.ndarray[real_t, ndim=1] buffer_arr = np.empty(size_buffer_as_int, dtype=C_real_t)
    tsyev(&V, &lo, &m, ptr_Ev, &m, ptr_ev, &buffer_arr[0], &size_buffer_as_int, &ignore)

    ### Note: the obtained eigen values are transposed at this point as they were passed
    ### assuming column-major (input was a symmetric matrix so it doesn't matter).
    ### If doing the rest of this part in C, would have to use tscal with incX given by
    ### the number of columns

    ### Now use the formula
    cdef size_t n = mu.shape[0]
    if isinstance(rng, np.random.Generator):
        sample_stdnorm = rng.standard_normal(size=(n, m), dtype=C_real_t)
    else:
        sample_stdnorm = rng.normal(size=(n, m)).astype(C_real_t)
    return mu + np.dot(
        sample_stdnorm,
        np.einsum("ij,i->ij", eig_Vecs, np.sqrt(1. / eig_vals))
    )

def add_to_diag(np.ndarray[real_t, ndim=2] XtX, bint add_bias, real_t lam):
    if not XtX.shape[0]:
        return None
    cdef size_t n = XtX.shape[0] - add_bias
    cdef size_t lda = XtX.shape[0]
    cdef size_t ix
    cdef real_t *ptr_XtX = &XtX[0,0]
    with nogil, boundscheck(False), nonecheck(False), wraparound(False):
        for ix in range(n):
            ptr_XtX[ix + ix*lda] += lam

def fill_lower_triangle(np.ndarray[real_t, ndim=2] XtX):
    if not XtX.shape[0]:
        return None
    cdef size_t n = XtX.shape[0]
    cdef size_t ix, row, col
    cdef real_t *ptr_XtX = &XtX[0,0]
    with nogil, boundscheck(False), nonecheck(False), wraparound(False):
        for row in range(1, n):
            for col in range(row):
                ptr_XtX[col + row*n] = ptr_XtX[row + col*n]

def solve_elasticnet(
        np.ndarray[real_t, ndim=2] XtX,
        np.ndarray[real_t, ndim=1] XtY2,
        np.ndarray[real_t, ndim=1] prev_coef,
        np.ndarray[real_t, ndim=1] coef,
        bint add_bias=1
    ):
    
    cdef int n = XtX.shape[0]
    cdef int ix
    cdef real_t *ptr_XtX = &XtX[0,0]
    cdef real_t *ptr_XtY_pos = &XtY2[0]
    cdef real_t *ptr_XtY_neg = ptr_XtY_pos + n
    cdef real_t *ptr_prev_pos = &prev_coef[0]
    cdef real_t *ptr_prev_neg = ptr_prev_pos + n
    cdef int one = 1

    cdef real_t diff_iter = 0.
    cdef real_t diff_val = 0.
    cdef real_t newval = 0

    with nogil, boundscheck(False), nonecheck(False), wraparound(False):
        while True:
            diff_iter = 0;
            for ix in range(n):
                newval = ptr_prev_pos[ix] + ptr_XtY_pos[ix] / ptr_XtX[ix + ix*n]
                newval = newval if newval >= 0. else 0.
                diff_val = newval - ptr_prev_pos[ix]
                if (fabs_t(diff_val) > 1e-10):
                    diff_iter += fabs_t(diff_val)
                    taxpy(&n, &diff_val, ptr_XtX + ix*n, &one, ptr_XtY_neg, &one)
                    diff_val = -diff_val
                    taxpy(&n, &diff_val, ptr_XtX + ix*n, &one, ptr_XtY_pos, &one)
                    ptr_prev_pos[ix] = newval;
            for ix in range(n):
                newval = ptr_prev_neg[ix] + ptr_XtY_neg[ix] / ptr_XtX[ix + ix*n]
                newval = newval if newval >= 0. else 0.
                diff_val = newval - ptr_prev_neg[ix]
                if fabs_t(diff_val) > 1e-10:
                    diff_iter += fabs_t(diff_val)
                    taxpy(&n, &diff_val, ptr_XtX + ix*n, &one, ptr_XtY_pos, &one)
                    diff_val = -diff_val
                    taxpy(&n, &diff_val, ptr_XtX + ix*n, &one, ptr_XtY_neg, &one)
                    ptr_prev_neg[ix] = newval
            
            if (isnan(diff_iter)) or (not isfinite(diff_iter)) or (diff_iter < 1e-8):
                break
        for ix in range(n):
            coef[ix] = ptr_prev_pos[ix] - ptr_prev_neg[ix]



from setuptools import setup, Extension
from Cython.Distutils import build_ext
import numpy as np
from sys import platform
import sys, os, subprocess, warnings, re

class build_ext_subclass( build_ext ):
    def build_extensions(self):
        if self.compiler.compiler_type == 'msvc':
            for e in self.extensions:
                e.extra_compile_args += ['/O2', '/openmp']
        else:
            if not self.check_for_variable_dont_set_march() and not self.check_cflags_contain_arch():
                self.add_march_native("c")
            if not self.check_for_variable_dont_set_march() and not self.check_cxxflags_contain_arch():
                self.add_march_native("c++")
            self.add_openmp_linkage() ### for c++ only

            for e in self.extensions:
                # e.extra_compile_args += ['-O3', '-march=native', '-fopenmp']
                # e.extra_link_args += ['-fopenmp']
                e.extra_compile_args += ['-O3']

                if e.language == "c++":
                    e.extra_compile_args += ['-std=c++11']

        build_ext.build_extensions(self)

    def check_cflags_contain_arch(self):
        if "CFLAGS" in os.environ:
            arch_list = ["-march", "-mcpu", "-mtune", "-msse", "-msse2", "-msse3", "-mssse3", "-msse4", "-msse4a", "-msse4.1", "-msse4.2", "-mavx", "-mavx2"]
            for flag in arch_list:
                if flag in os.environ["CFLAGS"]:
                    return True
        return False

    def check_cxxflags_contain_arch(self):
        if "CXXFLAGS" in os.environ:
            arch_list = ["-march", "-mcpu", "-mtune", "-msse", "-msse2", "-msse3", "-mssse3", "-msse4", "-msse4a", "-msse4.1", "-msse4.2", "-mavx", "-mavx2"]
            for flag in arch_list:
                if flag in os.environ["CXXFLAGS"]:
                    return True
        return False

    def check_for_variable_dont_set_march(self):
        return "DONT_SET_MARCH" in os.environ

    def add_march_native(self, lang):
        arg_march_native = "-march=native"
        arg_mcpu_native = "-mcpu=native"
        if self.test_supports_compile_arg(arg_march_native):
            for e in self.extensions:
                if (e.language is None and lang == "c") or (e.language == lang):
                    e.extra_compile_args.append(arg_march_native)
        elif self.test_supports_compile_arg(arg_mcpu_native):
            for e in self.extensions:
                if (e.language is None and lang == "c") or (e.language == lang):
                    e.extra_compile_args.append(arg_mcpu_native)

    def add_openmp_linkage(self):
        arg_omp1 = "-fopenmp"
        arg_omp2 = "-qopenmp"
        arg_omp3 = "-xopenmp"
        args_apple_omp = ["-Xclang", "-fopenmp", "-lomp"]
        args_apple_omp2 = ["-Xclang", "-fopenmp", "-L/usr/local/lib", "-lomp", "-I/usr/local/include"]
        if self.test_supports_compile_arg(arg_omp1, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args.append(arg_omp1)
                e.extra_link_args.append(arg_omp1)
        elif (sys.platform[:3].lower() == "dar") and self.test_supports_compile_arg(args_apple_omp, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-Xclang", "-fopenmp"]
                e.extra_link_args += ["-lomp"]
        elif (sys.platform[:3].lower() == "dar") and self.test_supports_compile_arg(args_apple_omp2, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-Xclang", "-fopenmp"]
                e.extra_link_args += ["-L/usr/local/lib", "-lomp"]
                e.include_dirs += ["/usr/local/include"]
        elif self.test_supports_compile_arg(arg_omp2, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args.append(arg_omp2)
                e.extra_link_args.append(arg_omp2)
        elif self.test_supports_compile_arg(arg_omp3, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args.append(arg_omp3)
                e.extra_link_args.append(arg_omp3)

    def test_supports_compile_arg(self, comm, with_omp=False):
        is_supported = False
        try:
            if not hasattr(self.compiler, "compiler_cxx"):
                return False
            if not isinstance(comm, list):
                comm = [comm]
            print("--- Checking compiler support for option '%s'" % " ".join(comm))
            fname = "contextualbandits_compiler_testing.cpp"
            with open(fname, "w") as ftest:
                ftest.write(u"int main(int argc, char**argv) {return 0;}\n")
            try:
                if not isinstance(self.compiler.compiler_cxx, list):
                    cmd = list(self.compiler.compiler_cxx)
                else:
                    cmd = self.compiler.compiler_cxx
            except Exception:
                cmd = self.compiler.compiler_cxx
            val_good = subprocess.call(cmd + [fname])
            if with_omp:
                with open(fname, "w") as ftest:
                    ftest.write(u"#include <omp.h>\nint main(int argc, char**argv) {return 0;}\n")
            try:
                val = subprocess.call(cmd + comm + [fname])
                is_supported = (val == val_good)
            except Exception:
                is_supported = False
        except Exception:
            pass
        try:
            os.remove(fname)
        except Exception:
            pass
        return is_supported

setup(
    name = 'contextualbandits',
    packages = ['contextualbandits', 'contextualbandits.linreg'],
    install_requires=[
        'numpy>=1.17',
        'scipy',
        'pandas>=0.25.0',
        'scikit-learn>=0.22',
        'joblib>=0.13',
        'cython'
    ],
    version = '0.3.18',
    description = 'Python Implementations of Algorithms for Contextual Bandits',
    author = 'David Cortes',
    author_email = 'david.cortes.rivera@gmail.com',
    url = 'https://github.com/david-cortes/contextualbandits',
    keywords = 'contextual bandits offset tree doubly robust policy linucb thompson sampling',
    classifiers = [],
    cmdclass = {'build_ext': build_ext_subclass},
    ext_modules = [
        Extension("contextualbandits.linreg._wrapper_double",
                  sources=["contextualbandits/linreg/linreg_double.pyx"],
                  include_dirs=[np.get_include()]),
        Extension("contextualbandits.linreg._wrapper_float",
                  sources=["contextualbandits/linreg/linreg_float.pyx"],
                  include_dirs=[np.get_include()]),
        Extension("contextualbandits._cy_utils", language="c++",
                  sources=["contextualbandits/_cy_utils.pyx"],
                  include_dirs=[np.get_include()])
    ]
)
